/* DIO-64 Service Module.

   Author: Reed Lai

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/* File level history (record changes for this file here.)

   v 0.3.0 18 Nov 2002 by Reed Lai
     Comments out the including of slab.h which is no longer used
     by this module.

     Uses slab.h in place of malloc.h.
     
   v 0.2.0 27 Feb 2002 by Reed Lai
     Use a new flag type for device information (ixisa_devinfo.)

   v 0.1.0 11 Jan 2002 by Reed Lai
     Rename DAQ to IXISA.

   v 0.0.0 24 Jul 2001 by Reed Lai
     create, blah blah... */

/* Mandatory */
#include <linux/kernel.h>		/* ta, kernel work */
#include <linux/module.h>		/* is a module */

/* Deal with CONFIG_MODVERSIONS that is defined in
   /usr/include/linux/config.h (config.h is included by module.h) */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

/* Additional */
#include <linux/fs.h>
	/* file system here */
#include <linux/wrapper.h>
	/* for compatibility with future version of Linux */

/* use I/O ports */
#include <asm/io.h>
#include <linux/ioport.h>

/* need kmalloc */
// #include <linux/slab.h>

/* irq service */
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <asm/irq.h>

/* Local */
#include "ixisa.h"
#include "_flag_inline.h"
#include "_dio64.h"

#define BUF_LEN 80

char msg[BUF_LEN];
char *msg_ptr;

ixisa_devinfo_t *dev;

/* symbols exportation */
// EXPORT_NO_SYMBOLS; /* do not export symbols */
EXPORT_SYMBOL_NOVERS(ixdio64_ioctl);
EXPORT_SYMBOL_NOVERS(ixdio64_release);
EXPORT_SYMBOL_NOVERS(ixdio64_open);

#if BOTTOM_HELF
void irq_handler_bh(void *data)
{
	/* bottom helf of irq handler
	 *
	 * Arguments:
	 *   data     data pointer as <linux/tqueue.h> defined
	 *
	 * Returned:  none */

	ixisa_devinfo_t *dp;

	dp = (ixisa_devinfo_t *) data;

	/* reserved */
}
#endif							/* BOTTOM_HELF */

void irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	/* Arguments:  as <linux/sched.h> defined
	 *
	 * Returned:  none */

#if BOTTOM_HELF
	static struct tq_struct task = { NULL, 0, irq_handler_bh, NULL };

	task.data = dev_id;
#endif							/* BOTTOM_HELF */

	send_sig(((ixisa_devinfo_t *) dev_id)->sig.sid,
			 ((ixisa_devinfo_t *) dev_id)->sig.task, 1);

	/* Scheduale bottom half to run */
#if BOTTOM_HELF
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
	queue_task(&task, &tq_immediate);
#else
	queue_task_irq(&task, &tq_immediate);
#endif
	mark_bh(IMMEDIATE_BH);
#endif							/* BOTTOM_HELF */

}

int set_signal(ixisa_signal_t * sig, ixisa_devinfo_t * dp)
{
	/* set user signal for interrupt
	 *
	 * Arguments:
	 *   sig      pointer to structure ixisa_signal for user signal
	 *   di       pointer to structure ixisa_devinfo for current device
	 *
	 * Returned:  SUCCESS or FAILURE */

	if (sig->sid) {
		/* has signal id, set */
		if (sig->pid) {
			/* user gives his process id */
			dp->sig.sid = sig->sid;
			dp->sig.pid = sig->pid;
			dp->sig.task = find_task_by_pid(sig->pid);
		} else if (sig->task) {
			/* user gives his task structure */
			dp->sig.sid = sig->sid;
			dp->sig.task = sig->task;
		} else {
			/* wrong! */
			return FAILURE;
		}
		dp->sig.is = sig->is;
		dp->sig.edge = sig->edge;
		dp->sig.bedge = sig->bedge;
	} else {
		/* no signal id, clear */
		dp->sig.sid = 0;
		dp->sig.pid = 0;
		dp->sig.task = 0;
		dp->sig.is = 0;
		dp->sig.edge = 0;
		dp->sig.bedge = 0;
	}
	return SUCCESS;
}

int write_reg(ixisa_reg_t * reg, unsigned long int base)
{
	/* Write to register
	 *
	 * Arguments:
	 *   reg      pointer to a structure ixisa_reg for register
	 *   base     base address for current device
	 *
	 * Returned:  SUCCESS or FAILURE */

	switch (reg->id) {
	case IXISA_DIO_A:			/* DIO 0-7 */
#if DIO_A_MASK == 0xff
		outb(((unsigned char) reg->value), base + DIO_A);
#else
		outb(((unsigned char) reg->value) & DIO_A_MASK, base + DIO_A);
#endif
		break;
	case IXISA_DIO_B:			/* DIO 8-15 */
#if DIO_B_MASK == 0xff
		outb(((unsigned char) reg->value), base + DIO_B);
#else
		outb(((unsigned char) reg->value) & DIO_B_MASK, base + DIO_B);
#endif
		break;
	case IXISA_DIO_C:			/* DIO 16-23 */
#if DIO_C_MASK == 0xff
		outb(((unsigned char) reg->value), base + DIO_C);
#else
		outb(((unsigned char) reg->value) & DIO_C_MASK, base + DIO_C);
#endif
		break;
	case IXISA_DIO_D:			/* DIO 24-31 */
#if DIO_D_MASK == 0xff
		outb(((unsigned char) reg->value), base + DIO_D);
#else
		outb(((unsigned char) reg->value) & DIO_D_MASK, base + DIO_D);
#endif
		break;
	case IXISA_8254C0:			/* 8254 Counter 0 */
#if _8254C0_MASK == 0xff
		outb(((unsigned char) reg->value), base + _8254C0);
#else
		outb(((unsigned char) reg->value) & _8254C0_MASK, base + _8254C0);
#endif
		break;
	case IXISA_8254C1:			/* 8254 Counter 1 */
#if _8254C1_MASK == 0xff
		outb(((unsigned char) reg->value), base + _8254C1);
#else
		outb(((unsigned char) reg->value) & _8254C1_MASK, base + _8254C1);
#endif
		break;
	case IXISA_8254C2:			/* 8254 Counter 2 */
#if _8254C2_MASK == 0xff
		outb(((unsigned char) reg->value), base + _8254C2);
#else
		outb(((unsigned char) reg->value) & _8254C2_MASK, base + _8254C2);
#endif
		break;
	case IXISA_8254CR:			/* 8254 Control Word */
#if _8254CR_MASK == 0xff
		outb(((unsigned char) reg->value), base + _8254CR);
#else
		outb(((unsigned char) reg->value) & _8254CR_MASK, base + _8254CR);
#endif
		break;
	case IXISA_8254C3:			/* 8254 Counter 3 */
#if _8254C3_MASK == 0xff
		outb(((unsigned char) reg->value), base + _8254C3);
#else
		outb(((unsigned char) reg->value) & _8254C3_MASK, base + _8254C3);
#endif
		break;
	case IXISA_8254C4:			/* 8254 Counter 4 */
#if _8254C4_MASK == 0xff
		outb(((unsigned char) reg->value), base + _8254C4);
#else
		outb(((unsigned char) reg->value) & _8254C4_MASK, base + _8254C4);
#endif
		break;
	case IXISA_8254C5:			/* 8254 Counter 5 */
#if _8254C5_MASK == 0xff
		outb(((unsigned char) reg->value), base + _8254C5);
#else
		outb(((unsigned char) reg->value) & _8254C5_MASK, base + _8254C5);
#endif
		break;
	case IXISA_8254CR1:			/* 8254-1 Control Word */
#if _8254CR1_MASK == 0xff
		outb(((unsigned char) reg->value), base + _8254CR1);
#else
		outb(((unsigned char) reg->value) & _8254CR1_MASK, base + _8254CR1);
#endif
		break;
	default:
		return FAILURE;
	}
	return SUCCESS;
}

int read_reg(ixisa_reg_t * reg, unsigned long int base)
{
	/* Read from register
	 *
	 * Arguments:
	 *   reg      pointer to structure ixisa_reg for register
	 *   base     base address for current device
	 *
	 * Returned:  SUCCESS or FAILURE */

	switch (reg->id) {
	case IXISA_DIO_A:			/* DIO 0-7 */
#if DIO_A_MASK == 0xff
		reg->value = inb(base + DIO_A);
#else
		reg->value = inb(base + DIO_A) & DIO_A_MASK;
#endif
		break;
	case IXISA_DIO_B:			/* DIO 8-15 */
#if DIO_B_MASK == 0xff
		reg->value = inb(base + DIO_B);
#else
		reg->value = inb(base + DIO_B) & DIO_B_MASK;
#endif
		break;
	case IXISA_DIO_C:			/* DIO 16-23 */
#if DIO_C_MASK == 0xff
		reg->value = inb(base + DIO_C);
#else
		reg->value = inb(base + DIO_C) & DIO_C_MASK;
#endif
		break;
	case IXISA_DIO_D:			/* DIO 24-31 */
#if DIO_D_MASK == 0xff
		reg->value = inb(base + DIO_D);
#else
		reg->value = inb(base + DIO_D) & DIO_D_MASK;
#endif
		break;
	case IXISA_8254C0:			/* 8254 Counter 0 */
#if _8254C0_MASK == 0xff
		reg->value = inb(base + _8254C0);
#else
		reg->value = inb(base + _8254C0) & _8254C0_MASK;
#endif
		break;
	case IXISA_8254C1:			/* 8254 Counter 1 */
#if _8254C1_MASK == 0xff
		reg->value = inb(base + _8254C1);
#else
		reg->value = inb(base + _8254C1) & _8254C1_MASK;
#endif
		break;
	case IXISA_8254C2:			/* 8254 Counter 2 */
#if _8254C2_MASK == 0xff
		reg->value = inb(base + _8254C2);
#else
		reg->value = inb(base + _8254C2) & _8254C2_MASK;
#endif
		break;
	case IXISA_8254CR:			/* 8254 Control Word */
#if _8254CR_MASK == 0xff
		reg->value = inb(base + _8254CR);
#else
		reg->value = inb(base + _8254CR) & _8254CR_MASK;
#endif
		break;
	case IXISA_8254C3:			/* 8254 Counter 3 */
#if _8254C3_MASK == 0xff
		reg->value = inb(base + _8254C3);
#else
		reg->value = inb(base + _8254C3) & _8254C3_MASK;
#endif
		break;
	case IXISA_8254C4:			/* 8254 Counter 4 */
#if _8254C4_MASK == 0xff
		reg->value = inb(base + _8254C4);
#else
		reg->value = inb(base + _8254C4) & _8254C4_MASK;
#endif
		break;
	case IXISA_8254C5:			/* 8254 Counter 5 */
#if _8254C5_MASK == 0xff
		reg->value = inb(base + _8254C5);
#else
		reg->value = inb(base + _8254C5) & _8254C5_MASK;
#endif
		break;
	case IXISA_8254CR1:			/* 8254-1 Control Word */
#if _8254CR1_MASK == 0xff
		reg->value = inb(base + _8254CR1);
#else
		reg->value = inb(base + _8254CR1) & _8254CR1_MASK;
#endif
		break;
	default:
		return FAILURE;
	}
	return SUCCESS;
}

ixisa_devinfo_t *align_to_minor(int minor)
{
	ixisa_devinfo_t *p;
	for (p = dev; p && p->no != minor && p->next_f; p = p->next_f);
	if (p->no == minor)
		return p;
	return 0;
}

int ixdio64_ioctl(struct inode *inode, struct file *file,
				  unsigned int ioctl_num, unsigned long ioctl_param)
{
	/* (export)
	 *
	 * This function is called by ixisa.o whenever a process tries to do and
	 * IO control on DIO device file
	 *
	 * Arguments: read <linux/fs.h> for (*ioctl) of struct file_operations
	 *
	 * Returned:  SUCCESS or FAILED */

	ixisa_devinfo_t *dp;

	dp = align_to_minor(MINOR(inode->i_rdev));

	if (!dp) {
		/* if not found, get out */
		return -EINVAL;
	}

	if (!dp->open) {
		/* if not yet opened, get out */
		return -EINVAL;
	}

	switch (ioctl_num) {
	case IXISA_GET_INFO:
		/* get device's information */
		ixisa_copy_devinfo((ixisa_devinfo_t *) ioctl_param, dp);
		break;
	case IXISA_SET_SIG:
		/* set signal for ixisa interrupt */
		if (set_signal((ixisa_signal_t *) ioctl_param, dp)) {
			return FAILURE;
		}
		break;
	case IXISA_READ_REG:
		/* read from register */
		if (read_reg((ixisa_reg_t *) ioctl_param, dp->base)) {
			return FAILURE;
		}
		break;
	case IXISA_WRITE_REG:
		/* write to register */
		if (write_reg((ixisa_reg_t *) ioctl_param, dp->base)) {
			return FAILURE;
		}
		break;
	default:
		return -EINVAL;
	}
	return SUCCESS;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
int ixdio64_release(struct inode *inode, struct file *file)
#else
void ixdio64_release(struct inode *inode, struct file *file)
#endif
{
	/* (export)
	 *
	 * This function is called by ixisa.o whenever a process attempts to
	 * closes the device file. It doesn't have a return value in kernel
	 * version 2.0.x because it can't fail (you must always be able to
	 * close a device).  In version 2.2.x it is allowed to fail.
	 *
	 * Arguments: read <linux/fs.h> for (*release) of struct file_operations
	 *
	 * Returned:  none */

	int minor;
	ixisa_devinfo_t *dp;

	minor = MINOR(inode->i_rdev);
	dp = align_to_minor(minor);

	if (!dp) {
		/* if not found, get out */
		return 0;
	}

	dp->open = 0;
	/* flag close */

	/* uninstall ISR */
	free_irq(dp->irq, dp);

	/* release io region */
	release_region(dp->base, IO_RANGE);

	MOD_DEC_USE_COUNT;

# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	return 0;
# endif
}

static struct file_operations fops = {
	open:ixdio64_open,
	release:ixdio64_release,
	ioctl:ixdio64_ioctl,
};

int ixdio64_open(struct inode *inode, struct file *file)
{
	/* (export)
	 *
	 * This function is called by ixisa.o whenever a process attempts to open
	 * the device file of DIO-64
	 *
	 * Arguments: read <linux/fs.h> for (*open) of struct file_operations
	 *
	 * Returned:  none */

	int minor;
	ixisa_devinfo_t *dp;
	ixisa_signal_t sig;

	minor = MINOR(inode->i_rdev);
	dp = align_to_minor(minor);

	if (!dp) {
		/* if not found, get out */
		return -EINVAL;
	}

	++(dp->open);
	if (dp->open > 1) {
		/* if already opened by someone, get out */
		--(dp->open);
		return -EBUSY;
	}

	/* clear signal setting */
	sig.sid = 0;
	sig.pid = 0;
	sig.task = 0;
	set_signal(&sig, dp);

	/* request io region */
	request_region(dp->base, IO_RANGE, MODULE_NAME);

	/* if irq is configured, install */
	if (dp->irq)
		request_irq(dp->irq, irq_handler, SA_INTERRUPT | SA_SHIRQ,
					dp->name, dp);

	MOD_INC_USE_COUNT;

	return SUCCESS;
}

void cleanup_module()
{
	/* cleanup this module
	 *
	 * Arguments: none
	 *
	 * Returned:  none */

	ixisa_devinfo_t *dp;

	printk(KERN_INFO FAMILY ": %s ", MODULE_NAME);

	/* uninstall file operations */
	for (dp = dev; dp; dp = dp->next_f) {
		dp->fops = 0;
		off_flag(&dp->flags, IXISA_FLAG_ATTEND);
		printk(".");
	}
	printk(" has been removed.\n");
}

int init_module()
{
	/* initialize this module
	 *
	 * Arguments: none
	 *
	 * Returned:
	 *   integer 0 means ok, otherwise failed (module can't be load) */

	ixisa_devinfo_t *dp;

	printk(KERN_INFO FAMILY ": %s ", MODULE_NAME);

	if (!ixisa_dev) {
		printk("device list is empty!\n");
		return FAILURE;
	}

	/* align to first DIO-64 in ixisa list */
	for (dev = ixisa_dev; dev && dev->id != DIO64; dev = dev->next);

	if (dev->id != DIO64) {
		printk("fail!\n");
		return FAILURE;
	}

	/* install file operations */
	for (dp = dev; dp; dp = dp->next_f) {
		dp->fops = &fops;
		on_flag(&dp->flags, IXISA_FLAG_ATTEND);
		printk(".");
	}

	printk(" is ready.\n");

	return SUCCESS;
}
