/* ISO-C64 Service Module.

   Author: Reed Lai

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/* File level history (record changes for this file here.)

   v 0.2.0 18 Nov 2002 by Reed Lai
     Comments out the including of slab.h which is no longer used
     by this module.

     Uses the slab.h in place of malloc.h.

   v 0.1.0 27 Feb 2002 by Reed Lai
     Use a new flag type for device information (ixisa_devinfo.)

   v 0.0.0 16 Jan 2002 by Reed Lai
     Create. */

/* Mandatory */
#include <linux/kernel.h>		/* ta, kernel work */
#include <linux/module.h>		/* is a module */

/* Deal with CONFIG_MODVERSIONS that is defined in
   /usr/include/linux/config.h (config.h is included by module.h) */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

/* Additional */
#include <linux/fs.h>
	/* file system here */
#include <linux/wrapper.h>
	/* for compatibility with future version of Linux */

/* use I/O ports */
#include <asm/io.h>
#include <linux/ioport.h>

/* need kmalloc */
// #include <linux/slab.h>

/* irq service */
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <asm/irq.h>

/* Local */
#include "ixisa.h"
#include "_flag_inline.h"
#include "_isoc64.h"

#define BUF_LEN 80

char msg[BUF_LEN];
char *msg_ptr;

ixisa_devinfo_t *dev;

/* symbols exportation */
// EXPORT_NO_SYMBOLS; /* do not export symbols */
EXPORT_SYMBOL_NOVERS(ixisoc64_ioctl);
EXPORT_SYMBOL_NOVERS(ixisoc64_release);
EXPORT_SYMBOL_NOVERS(ixisoc64_open);

#if BOTTOM_HELF
void irq_handler_bh(void *data)
{
	/* bottom helf of irq handler
	 *
	 * Arguments:
	 *   data     data pointer as <linux/tqueue.h> defined
	 *
	 * Returned:  none */

	ixisa_devinfo_t *dp;

	dp = (ixisa_devinfo_t *) data;

	/* reserved */
}
#endif							/* BOTTOM_HELF */

void irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	/* Arguments:  as <linux/sched.h> defined
	 *
	 * Returned:  none */

#if BOTTOM_HELF
	static struct tq_struct task = { NULL, 0, irq_handler_bh, NULL };

	task.data = dev_id;
#endif							/* BOTTOM_HELF */

	send_sig(((ixisa_devinfo_t *) dev_id)->sig.sid,
			 ((ixisa_devinfo_t *) dev_id)->sig.task, 1);

	/* Scheduale bottom half to run */
#if BOTTOM_HELF
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
	queue_task(&task, &tq_immediate);
#else
	queue_task_irq(&task, &tq_immediate);
#endif
	mark_bh(IMMEDIATE_BH);
#endif							/* BOTTOM_HELF */
}

int set_signal(ixisa_signal_t * sig, ixisa_devinfo_t * dp)
{
	/* set user signal for interrupt
	 *
	 * Arguments:
	 *   sig      pointer to structure ixisa_signal for user signal
	 *   di       pointer to structure ixisa_devinfo for current device
	 *
	 * Returned:  SUCCESS or FAILURE */

	if (sig->sid) {
		/* has signal id, set */
		if (sig->pid) {
			/* user gives his process id */
			dp->sig.sid = sig->sid;
			dp->sig.pid = sig->pid;
			dp->sig.task = find_task_by_pid(sig->pid);
		} else if (sig->task) {
			/* user gives his task structure */
			dp->sig.sid = sig->sid;
			dp->sig.task = sig->task;
		} else {
			/* wrong! */
			return FAILURE;
		}
		dp->sig.is = sig->is;
		dp->sig.edge = sig->edge;
		dp->sig.bedge = sig->bedge;
	} else {
		/* no signal id, clear */
		dp->sig.sid = 0;
		dp->sig.pid = 0;
		dp->sig.task = 0;
		dp->sig.is = 0;
		dp->sig.edge = 0;
		dp->sig.bedge = 0;
	}
	return SUCCESS;
}

int write_reg(ixisa_reg_t * reg, unsigned long int base)
{
	/* Write to register
	 *
	 * Arguments:
	 *   reg      pointer to a structure ixisa_reg for register
	 *   base     base address for current device
	 *
	 * Returned:  SUCCESS or FAILURE */
	switch (reg->id) {
	case IXISA_DO_A:			/* DO Channel 0-7 */
#if DO_A_MASK == 0xff
		outb(((unsigned char) reg->value), base + DO_A);
#else
		outb(((unsigned char) reg->value) & DO_A_MASK, base + DO_A);
#endif
		break;
	case IXISA_DO_B:			/* DO Channel 8-15 */
#if DO_B_MASK == 0xff
		outb(((unsigned char) reg->value), base + DO_B);
#else
		outb(((unsigned char) reg->value) & DO_B_MASK, base + DO_B);
#endif
		break;
	case IXISA_DO_C:			/* DO Channel 16-23 */
#if DO_C_MASK == 0xff
		outb(((unsigned char) reg->value), base + DO_C);
#else
		outb(((unsigned char) reg->value) & DO_C_MASK, base + DO_C);
#endif
		break;
	case IXISA_DO_D:			/* DO Channel 24-31 */
#if DO_D_MASK == 0xff
		outb(((unsigned char) reg->value), base + DO_D);
#else
		outb(((unsigned char) reg->value) & DO_D_MASK, base + DO_D);
#endif
		break;
	case IXISA_DO_E:			/* DO Channel 32-39 */
#if DO_E_MASK == 0xff
		outb(((unsigned char) reg->value), base + DO_E);
#else
		outb(((unsigned char) reg->value) & DO_E_MASK, base + DO_E);
#endif
		break;
	case IXISA_DO_F:			/* DO Channel 40-47 */
#if DO_F_MASK == 0xff
		outb(((unsigned char) reg->value), base + DO_F);
#else
		outb(((unsigned char) reg->value) & DO_F_MASK, base + DO_F);
#endif
		break;
	case IXISA_DO_G:			/* DO Channel 48-55 */
#if DO_G_MASK == 0xff
		outb(((unsigned char) reg->value), base + DO_G);
#else
		outb(((unsigned char) reg->value) & DO_G_MASK, base + DO_G);
#endif
		break;
	case IXISA_DO_H:			/* DO Channel 56-63 */
#if DO_H_MASK == 0xff
		outb(((unsigned char) reg->value), base + DO_H);
#else
		outb(((unsigned char) reg->value) & DO_H_MASK, base + DO_H);
#endif
		break;
	default:
		return FAILURE;
	}
	return SUCCESS;
}

int read_reg(ixisa_reg_t * reg, unsigned long int base)
{
	/* Read from register
	 *
	 * Arguments:
	 *   reg      pointer to structure ixisa_reg for register
	 *   base     base address for current device
	 *
	 * Returned:  SUCCESS or FAILURE */
	switch (reg->id) {
	default:
		return FAILURE;
	}
	return SUCCESS;
}

ixisa_devinfo_t *align_to_minor(int minor)
{
	ixisa_devinfo_t *dp;
	for (dp = dev; dp && dp->no != minor && dp->next_f; dp = dp->next_f);
	if (dp->no == minor)
		return dp;
	return 0;
}

int ixisoc64_ioctl(struct inode *inode, struct file *file,
				   unsigned int ioctl_num, unsigned long ioctl_param)
{
	/* (export)
	 *
	 * This function is called by ixisa.o whenever a process tries to do an
	 * IO control on IXISA device file
	 *
	 * Arguments: read <linux/fs.h> for (*ioctl) of struct file_operations
	 *
	 * Returned:  SUCCESS or FAILED */

	ixisa_devinfo_t *dp;

	dp = align_to_minor(MINOR(inode->i_rdev));

	if (!dp) {
		/* if not found, get out */
		return -EINVAL;
	}

	if (!dp->open) {
		/* if not yet opened, get out */
		return -EINVAL;
	}

	switch (ioctl_num) {
	case IXISA_GET_INFO:
		/* get device's information */
		ixisa_copy_devinfo((ixisa_devinfo_t *) ioctl_param, dp);
		break;
	case IXISA_SET_SIG:
		/* set signal for ixisa interrupt */
		  
		if (set_signal((ixisa_signal_t *) ioctl_param, dp)) {
			return FAILURE;
		}
		break;
	case IXISA_READ_REG:
		/* read from register */
		if (read_reg((ixisa_reg_t *) ioctl_param, dp->base)) {
			return FAILURE;
		}
		break;
	case IXISA_WRITE_REG:
		/* write to register */
		if (write_reg((ixisa_reg_t *) ioctl_param, dp->base)) {
			return FAILURE;
		}
		break;
	default:
		return -EINVAL;
	}
	return SUCCESS;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
int ixisoc64_release(struct inode *inode, struct file *file)
#else
void ixisoc64_release(struct inode *inode, struct file *file)
#endif
{
	/* (export)
	 *
	 * This function is called by ixisa.o whenever a process attempts to
	 * closes the device file. It doesn't have a return value in kernel
	 * version 2.0.x because it can't fail (you must always be able to
	 * close a device).  In version 2.2.x it is allowed to fail.
	 *
	 * Arguments: read <linux/fs.h> for (*release) of struct file_operations
	 *
	 * Returned:  none */

	int minor;
	ixisa_devinfo_t *dp;

	minor = MINOR(inode->i_rdev);
	dp = align_to_minor(minor);

	if (!dp) {
		/* if not found, get out */
		return 0;
	}

	dp->open = 0;
	/* flag close */

	/* uninstall ISR */
	free_irq(dp->irq, dp);

	/* release io region */
	release_region(dp->base, IO_RANGE);

	MOD_DEC_USE_COUNT;

# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	return 0;
# endif
}

static struct file_operations fops = {
	open:ixisoc64_open,
	release:ixisoc64_release,
	ioctl:ixisoc64_ioctl,
};

int ixisoc64_open(struct inode *inode, struct file *file)
{
	/* (export)
	 *
	 * This function is called by ixisa.o whenever a process attempts to open
	 * the device file of ISO-C64
	 *
	 * Arguments: read <linux/fs.h> for (*open) of struct file_operations
	 *
	 * Returned:  none */

	int minor;
	ixisa_devinfo_t *dp;
	ixisa_signal_t sig;

	minor = MINOR(inode->i_rdev);
	dp = align_to_minor(minor);

	if (!dp) {
		/* if not found, get out */
		return -EINVAL;
	}

	++(dp->open);
	if (dp->open > 1) {
		/* if already opened by someone, get out */
		--(dp->open);
		return -EBUSY;
	}

	/* clear signal setting */
	sig.sid = 0;
	sig.pid = 0;
	sig.task = 0;
	set_signal(&sig, dp);

	/* request io region */
	request_region(dp->base, IO_RANGE, MODULE_NAME);

	/* if irq is configured, install */
	if (dp->irq)
		request_irq(dp->irq, irq_handler, SA_INTERRUPT | SA_SHIRQ,
					dp->name, dp);

	MOD_INC_USE_COUNT;

	return SUCCESS;
}

void cleanup_module()
{
	/* cleanup this module
	 *
	 * Arguments: none
	 *
	 * Returned:  none */

	ixisa_devinfo_t *dp;

	printk(KERN_INFO FAMILY ": %s ", MODULE_NAME);

	/* uninstall file operations */
	for (dp = dev; dp; dp = dp->next_f) {
		dp->fops = 0;
		off_flag(&dp->flags, IXISA_FLAG_ATTEND);
		printk(".");
	}
	printk(" has been removed.\n");
}

int init_module()
{
	/* initialize this module
	 *
	 * Arguments: none
	 *
	 * Returned:
	 *   integer 0 means ok, otherwise failed (module can't be load) */

	ixisa_devinfo_t *dp;

	printk(KERN_INFO FAMILY ": %s ", MODULE_NAME);

	if (!ixisa_dev) {
		printk("device list is empty!\n");
		return FAILURE;
	}

	/* align to first ISO-C64 in ixisa list */
	for (dev = ixisa_dev; dev && dev->id != ISOC64; dev = dev->next);

	if (dev->id != ISOC64) {
		printk("fail!\n");
		return FAILURE;
	}

	/* install file operations */
	for (dp = dev; dp; dp = dp->next_f) {
		dp->fops = &fops;
		on_flag(&dp->flags, IXISA_FLAG_ATTEND);
		printk(".");
	}

	printk(" is ready.\n");

	return SUCCESS;
}
