/* STEP-200 service module.

   Author: Reed Lai

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/* File level history (record changes for this file here.)

  v 0.6.0 18 Nov 2002 by Reed Lai
    Comments out the including of the slab.h which is no longer used
    by this module.

    Uses slab.h in place of malloc.h.
    Corrects some register names.
  v 0.5.1 27 Feb 2002 by Reed Lai
    Use a new flag type for device information (ixisa_devinfo.)

  v 0.4.1 29 Jan 2002 by Reed Lai
    Refine some register names.

  v 0.3.1 11 Jan 2002 by Reed Lai
    Rename DAQ to IXISA.

  v 0.2.1 24 Jul 2001 by Reed Lai
    Strip codes of the Bottom Helf of Interrupt Handler by flag
	BOTTOM_HELF
	Rename the short name of Control Word Register from CWR to CR

  v 0.0.1 8 May 2001 by Reed Lai
    Correct the removing family fop of cleanup_module by "next" to "next_f"
    This bug will cause the cleanup_module to remove fop that does not belong
    to it.

  v 0.0.0 19 Feb 2001 by Reed Lai
    create, blah blah... */

/* Mandatory */
#include <linux/kernel.h>		/* ta, kernel work */
#include <linux/module.h>		/* is a module */

/* Deal with CONFIG_MODVERSIONS that is defined in
   /usr/include/linux/config.h (config.h is included by module.h) */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

/* Additional */
#include <linux/fs.h>
	/* file system here */
#include <linux/wrapper.h>
	/* for compatibility with future version of Linux */

/* use I/O ports */
#include <asm/io.h>
#include <linux/ioport.h>

/* need kmalloc */
// #include <linux/slab.h>

/* irq service */
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <asm/irq.h>

/* Local */
#include "ixisa.h"
#include "_flag_inline.h"
#include "_step200.h"

#define BUF_LEN 80

char msg[BUF_LEN];
char *msg_ptr;

ixisa_devinfo_t *dev;

/* symbols exportation */
// EXPORT_NO_SYMBOLS; /* do not export symbols */
EXPORT_SYMBOL_NOVERS(ixstep200_ioctl);
EXPORT_SYMBOL_NOVERS(ixstep200_release);
EXPORT_SYMBOL_NOVERS(ixstep200_open);

int write_reg(ixisa_reg_t * reg, unsigned long int base)
{
	/* Write to register
	 *
	 * Arguments:
	 *   reg      pointer to a structure ixisa_reg for register
	 *   base     base address for current device
	 *
	 * Returned:  SUCCESS or FAILURE */


	switch (reg->id) {
	case IXISA_FIFO:
		/* FIFO Register */
#if FIFO_MASK == 0xff
		outb(((unsigned char) reg->value), base + FIFO);
#else
		outb(((unsigned char) reg->value) & FIFO_MASK, base + FIFO);
#endif
		break;
	case IXISA_RR:
		/* RSTFF Register */
#if RR_MASK == 0xff
		outb(((unsigned char) reg->value), base + RR);
#else
		outb(((unsigned char) reg->value) & RR_MASK, base + RR);
#endif
		break;
	default:
		return FAILURE;
	}
	return SUCCESS;
}

int read_reg(ixisa_reg_t * reg, unsigned long int base)
{
	/* Read from register
	 *
	 * Arguments:
	 *   reg      pointer to structure ixisa_reg for register
	 *   base     base address for current device
	 *
	 * Returned:  SUCCESS or FAILURE */

	switch (reg->id) {
	case IXISA_DI1:
		/* DI1 Register, original point switch of X-axis, low active */
#if DI1_MASK == 0xff
		reg->value = inb(base + DI1);
#else
		reg->value = inb(base + DI1) & DI1_MASK;
#endif
		break;
	case IXISA_DI2:
		/* DI2 Register, original point switch of Y-axis, low active */
#if DI2_MASK == 0xff
		reg->value = inb(base + DI2);
#else
		reg->value = inb(base + DI2) & DI2_MASK;
#endif
		break;
	case IXISA_STS:
		/* STS Register */
#if STS_MASK == 0xff
		reg->value = inb(base + STS);
#else
		reg->value = inb(base + STS) & STS_MASK;
#endif
		break;
	default:
		return FAILURE;
	}
	return SUCCESS;
}

ixisa_devinfo_t *align_to_minor(int minor)
{
	ixisa_devinfo_t *dp;
	for (dp = dev; dp && dp->no != minor && dp->next_f; dp = dp->next_f);
	if (dp->no == minor)
		return dp;
	return 0;
}

int ixstep200_ioctl(struct inode *inode, struct file *file,
				 unsigned int ioctl_num, unsigned long ioctl_param)
{
	/* (export)
	 *
	 * This function is called by ixisa.o whenever a process tries to do and
	 * IO control on PIO device file
	 *
	 * Arguments: read <linux/fs.h> for (*ioctl) of struct file_operations
	 *
	 * Returned:  SUCCESS or FAILED */

	ixisa_devinfo_t *dp;

	dp = align_to_minor(MINOR(inode->i_rdev));

	if (!dp) {
		/* if not found, get out */
		return -EINVAL;
	}

	if (!dp->open) {
		/* if not yet opened, get out */
		return -EINVAL;
	}

	switch (ioctl_num) {
	case IXISA_GET_INFO:
		/* get device's information */
		ixisa_copy_devinfo((ixisa_devinfo_t *) ioctl_param, dp);
		break;
	case IXISA_READ_REG:
		/* read from register */
		if (read_reg((ixisa_reg_t *) ioctl_param, dp->base)) {
			return FAILURE;
		}
		break;
	case IXISA_WRITE_REG:
		/* write to register */
		if (write_reg((ixisa_reg_t *) ioctl_param, dp->base)) {
			return FAILURE;
		}
		break;
	default:
		return -EINVAL;
	}
	return SUCCESS;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
int ixstep200_release(struct inode *inode, struct file *file)
#else
void ixstep200_release(struct inode *inode, struct file *file)
#endif
{
	/* (export)
	 *
	 * This function is called by ixisa.o whenever a process attempts to
	 * closes the device file. It doesn't have a return value in kernel
	 * version 2.0.x because it can't fail (you must always be able to
	 * close a device).  In version 2.2.x it is allowed to fail.
	 *
	 * Arguments: read <linux/fs.h> for (*release) of struct file_operations
	 *
	 * Returned:  none */

	int minor;
	ixisa_devinfo_t *dp;

	minor = MINOR(inode->i_rdev);
	dp = align_to_minor(minor);

	if (!dp) {
		/* if not found, get out */
		return 0;
	}

	dp->open = 0;
	/* flag close */


	/* release io region */
	release_region(dp->base, IO_RANGE);

	MOD_DEC_USE_COUNT;

# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	return 0;
# endif
}

static struct file_operations fops = {
	open:ixstep200_open,
	release:ixstep200_release,
	ioctl:ixstep200_ioctl,
};

int ixstep200_open(struct inode *inode, struct file *file)
{
	/* (export)
	 * This function is called by ixisa.o whenever a process attempts to open
	 * the device file of IXISA-STEP200
	 *
	 * Arguments: read <linux/fs.h> for (*open) of struct file_operations
	 *
	 * Returned:  none */

	int minor;
	ixisa_devinfo_t *dp;

	minor = MINOR(inode->i_rdev);
	dp = align_to_minor(minor);

	if (!dp) {
		/* if not found, get out */
		return -EINVAL;
	}

	++(dp->open);
	if (dp->open > 1) {
		/* if already opened by someone, get out */
		--(dp->open);
		return -EBUSY;
	}

	/* request io region */
	request_region(dp->base, IO_RANGE, MODULE_NAME);


	MOD_INC_USE_COUNT;

	return SUCCESS;
}

void cleanup_module()
{
	ixisa_devinfo_t *dp;

	printk(KERN_INFO FAMILY ": %s ", MODULE_NAME);

	/* uninstall file operations */
	for (dp = dev; dp; dp = dp->next_f) {
		dp->fops = 0;
		off_flag(&dp->flags, IXISA_FLAG_ATTEND);
		printk(".");
	}
	printk(" has been removed.\n");
}

int init_module()
{
	/* initialize this module
	 *
	 * Arguments: none
	 *
	 * Returned:
	 *   integer 0 means ok, otherwise failed (module can't be load) */

	ixisa_devinfo_t *dp;

	printk(KERN_INFO FAMILY ": %s... ", MODULE_NAME);

	if (!ixisa_dev) {
		printk("device list is empty!\n");
		return FAILURE;
	}

	/* align to first IXISA-STEP200 in ixisa list */
	for (dev = ixisa_dev; dev && dev->id != STEP200; dev = dev->next);

	if (!dev) {
		printk("no %s card registered!\n", STEP200_NAME);
		return FAILURE;
	}

	/* install file operations */
	for (dp = dev; dp; dp = dp->next_f) {
		dp->fops = &fops;
		on_flag(&dp->flags, IXISA_FLAG_ATTEND);
		printk(".");
	}

	printk(" ready.\n");

	return SUCCESS;
}
